/*
 * main.c
 *
 *  Created on: Apr 5, 2011
 *      Author: alex copot
 *      Email: alex.mihai.c@gmail.com
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include<util/delay.h>

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#define _DEBUG 1

#ifdef _DEBUG
int uart_putchar(char c, FILE *unused);
int uart_getchar(FILE* f);
FILE mystdout = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);
FILE mystdin =  FDEV_SETUP_STREAM(NULL, uart_getchar, _FDEV_SETUP_READ);
#endif

#define M_LEFT		1
#define M_RIGHT		2

#define FORWARD		1
#define BACKWARD	2
#define STOP		3

#define MOTOR_DDPORT DDRD
#define MOTOR_DPORT PORTD

#define MOTOR_RIGHT1 _BV(PD2)
#define MOTOR_RIGHT2 _BV(PD4)

#define MOTOR_LEFT1 _BV(PD3)
#define MOTOR_LEFT2 _BV(PD5)

#define TOP 15

#define AVG_THRESHOLD   50
#define DELTA_THRESHOLD 30

void init_hbridge()
{
	// FastPWM A si B
	TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(WGM11);
	TCCR1B =  _BV(WGM13) | _BV(WGM12) |  _BV(CS12) | _BV(CS10);
	
	MOTOR_DDPORT |= MOTOR_LEFT1 | MOTOR_RIGHT1 | MOTOR_LEFT2 | MOTOR_RIGHT2;

	OCR1A = 8;
	OCR1B = 8;
	ICR1 = TOP;
}
/* Generator de sunet */
void init_sound()
{
	DDRD |= _BV(PD7);
	TCCR2 =   _BV(WGM21) | _BV(WGM20) |  _BV(CS22) | _BV(CS20)  | _BV(COM21);
	OCR2 = 200;
}

void init_sensors()
{
	ADMUX |= _BV(REFS0);
	ADCSRA |= _BV(ADEN) | _BV(ADATE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);
}

int sensor_get(uint8_t channel)
{
	ADMUX &= 0xe0;
	ADMUX |= channel;
	ADCSRA |= _BV(ADSC) | _BV(ADEN);
	while ( (ADCSRA & _BV(ADIF)) == 0 );

	ADCSRA &= ~_BV(ADEN);

	return ADC;
}
#ifdef _DEBUG
void USART_init(void)
{
	/* 9600 bps la 16 Mhz */
	UBRRH = 0;
	UBRRL = 103;

	/* 8biti date */
	UCSRC = _BV(URSEL) | _BV(UCSZ1) | _BV(UCSZ0);

	/* activeaza RX, TX */
	UCSRB = _BV(RXEN) | _BV(TXEN);
}

int uart_putchar(char c, FILE *unused)
{
	if (c == '\n')
		uart_putchar('\r', 0);
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = c;
	return 0;
}

// Functie ce trateaza primirea unui caracter pe seriala
int uart_getchar(FILE* f)
{
	char c;

    loop_until_bit_is_set(UCSRA, RXC);
    c = UDR;
    if(c=='\r')
		c='\n';
    uart_putchar(c,NULL);
    return c;
}
#endif

/* Configureaza timerul 1 pentru motorul @motor_idx, setand
 * directia si viteaza de miscare */
void setRotation(int motor_idx, int direction, int speed)
{
	if (direction == FORWARD) {
		switch (motor_idx) {
			case M_LEFT:
				MOTOR_DPORT &= ~MOTOR_LEFT1;
				OCR1A = speed;
				break;
			case M_RIGHT:
				MOTOR_DPORT &= ~MOTOR_RIGHT1;
				OCR1B = speed;
				break;
		}
	} else
		if (direction == BACKWARD) {
			switch (motor_idx) {
				case M_LEFT:
					MOTOR_DPORT |= MOTOR_LEFT1;
					OCR1A = TOP - speed;
					break;
				case M_RIGHT:
					MOTOR_DPORT |= MOTOR_RIGHT1;
					OCR1B = TOP - speed;
					break;
			}
		} else {
			switch (motor_idx) {
				case M_LEFT:
					MOTOR_DPORT |= MOTOR_LEFT1 | MOTOR_LEFT2;
					OCR1A = 0;
					break;
				case M_RIGHT:
					MOTOR_DPORT |= MOTOR_RIGHT1 | MOTOR_RIGHT2;
					OCR1B = 0;
					break;
			}
		}
}

#ifdef TEST
void test_setRotation()
{
	int i;

	init_hbridge();
	OCR1A = 4;
	OCR1B = 4;
	USART_init();

	stdin = &mystdin;                                                                                                                                        
	stdout = &mystdout;
	i = ICR1 - 1;
	while (i < ICR1) {
		setRotation(M_LEFT, FORWARD, i);
		printf("left f\n");
		for (i = 0; i < 200; i++)
			_delay_ms(16);
		setRotation(M_LEFT, BACKWARD, i);
		printf("left b\n");
		for (i = 0; i < 200; i++)
			_delay_ms(16);

		setRotation(M_RIGHT, FORWARD, i);
		printf("right f\n");
		for (i = 0; i < 200; i++)
			_delay_ms(16);
		printf("right b\n");
		setRotation(M_RIGHT, BACKWARD, i);
		for (i = 0; i < 200; i++)
			_delay_ms(16);
		printf("\n");
		i++;
	}
}
#endif

int main()
{
	int spdL, spdR, i;
	int left, right, delta, avg;
	unsigned int count = 0;
	unsigned int tcnt = 0;

#ifdef TEST
	test_setRotation();
	
	MOTOR_DPORT = MOTOR_RIGHT1;
	MOTOR_DPORT |= MOTOR_LEFT2 | MOTOR_RIGHT2;

	for (i = 0; i < 200; i++)
		_delay_ms(16);
	MOTOR_DPORT  = MOTOR_LEFT1;
	MOTOR_DPORT |= MOTOR_RIGHT2 | MOTOR_LEFT2;
#endif

	
	init_hbridge();
	init_sensors();
	init_sound();

#ifdef _DEBUG
	USART_init();
	stdin = &mystdin;
	stdout = &mystdout;
#endif

	PORTD |= _BV(PD6);
	DDRC |= _BV(PC0);

	avg = 100;
	while (1) {
		
		count++;
		if (count >= 750 - 3 * avg) {
			count = 0;
			 TCCR2 ^= _BV(COM21);
		}

		tcnt++;
		if ((tcnt > 10000) && ((PIND & _BV(PD6)) == 0) ) {
			PORTC ^= _BV(PC0);
			tcnt = 0;
		}

		left = sensor_get(0);
		right  = sensor_get(1);

		delta = left - right;
		avg = (left + right) / 2;

		if ( delta > DELTA_THRESHOLD && avg > AVG_THRESHOLD) {
			spdL = 10;
			spdR = 5;
		} else
		if ( delta < -DELTA_THRESHOLD && avg > AVG_THRESHOLD) {
			spdL = 5;
			spdR = 10;
		} else {
			if ( left > 250 && right > 250 ) {

				setRotation(M_LEFT, STOP, 0);
				setRotation(M_RIGHT, STOP, 0);
				/* back */
				setRotation(M_LEFT, BACKWARD, 8);
				setRotation(M_RIGHT, BACKWARD, 8);
#ifdef _DEBUG
				printf("back\n");
#endif
				for (i = 0; i < 120; i++) {
					_delay_ms(10);
				}
				/* rotate in place */
				setRotation(M_LEFT, FORWARD, 8);
				setRotation(M_RIGHT, BACKWARD, 8);
#ifdef _DEBUG
				printf("rotate\n");
#endif
				for (i = 0; i < 50; i++) {
					_delay_ms(10);
				}

				setRotation(M_LEFT, STOP, 0);
				setRotation(M_RIGHT, STOP, 0);
#ifdef _DEBUG
				printf("forward\n");
#endif
				setRotation(M_LEFT, FORWARD, 8);
				setRotation(M_RIGHT, FORWARD, 8);
				spdL = spdR = 8;
			} else {
			   spdL = 8;
			   spdR = 8;
			}
		}
		setRotation(M_LEFT, FORWARD, spdL);
		setRotation(M_RIGHT, FORWARD, spdR);
	}

	return 0;
}


